from example_app.comp.beans import SomeBean
from example_app.model.my_dto import MyDto
from jivago.lang.annotations import Inject
from jivago.wsgi.annotations import Resource, Path
from jivago.wsgi.methods import GET, POST, DELETE
from jivago.wsgi.request import Request
from jivago.wsgi.response import Response


@Resource("/hello")
class HelloWorldResource(object):

    @Inject
    def __init__(self, some_bean: SomeBean):
        self.some_bean = some_bean

    @GET
    def get_hello(self) -> str:
        return self.some_bean.say_hello()

    @POST
    @Path("/{name}")
    def post_hello(self, name: str) -> str:
        print("name: {}".format(name))
        return self.some_bean.say_hello()

    @Path("/delete")
    @DELETE
    def delete_hello(self) -> str:
        raise NotImplementedError("delete this")

    @Path("/request")
    @GET
    @POST
    def read_request(self, request: Request) -> Response:
        print(request)
        return Response(402, {}, str(request))

    @Path("/request/json")
    @POST
    def read_request_body_from_dict(self, body: dict) -> str:
        return str(body)

    @Path("/request/dto")
    @POST
    def read_with_dto(self, a_dto: MyDto) -> str:
        return str(a_dto)

    @GET
    @Path("/query")
    def with_query(self, name: str) -> str:
        return "Hello {}!".format(name)



class Headers(object):

    def __init__(self, content=None):
        self.content = {} if content is None else content

    def __setitem__(self, key: str, value: str):
        self.content[key.upper()] = value

    def __getitem__(self, item: str) -> str:
        return self.content.get(item.upper())

    def values(self):
        return self.content.values()

    def keys(self):
        return self.content.keys()

    def items(self):
        return self.content.items()



import collections
from typing import Any, _Union, TypingMeta

from jivago.inject.registry import Registry
from jivago.lang.annotations import Serializable
from jivago.lang.stream import Stream
from jivago.wsgi.incorrect_attribute_type_exception import IncorrectAttributeTypeException


class DtoSerializationHandler(object):

    def __init__(self, registry: Registry, root_package: str):
        self.root_package = root_package
        self.registry = registry

    def is_serializable(self, clazz: type) -> bool:
        return self.registry.is_annotated(clazz, Serializable)

    def deserialize(self, body: dict, clazz: type) -> Any:
        constructor = clazz.__init__
        if constructor == object.__init__:
            return self.__reflexively_inject_attributes(clazz, body)
        else:
            return self.__inject_constructor(clazz, constructor, body)

    def serialize(self, dto):
        if isinstance(dto, list):
            return [self.serialize(x) for x in dto]
        if self.is_serializable(dto.__class__):
            dictionary = dto.__dict__
            for key, value in dictionary.items():
                if self.is_serializable(value.__class__) or isinstance(value, list):
                    dictionary[key] = self.serialize(value)
            return dictionary
        return dto

    def __inject_constructor(self, clazz, constructor, body):
        the_object = object.__new__(clazz)
        parameters = []

        parameter_declarations = constructor.__annotations__.items()

        for attribute, declared_type in parameter_declarations:
            if attribute == 'return':
                break
            allowed_attribute_types = [declared_type] if not isinstance(declared_type,
                                                                        _Union) else declared_type.__args__

            if isinstance(declared_type, TypingMeta) and declared_type.__name__ == 'List':
                parameters.append([self.deserialize(body[attribute][x], declared_type.__args__[0]) for x in
                                   range(0, len(body[attribute]))])
            elif Stream(allowed_attribute_types).anyMatch(
                    lambda attribute_type: isinstance(body.get(attribute), attribute_type)):
                parameters.append(body.get(attribute))
            else:
                raise IncorrectAttributeTypeException(attribute, declared_type)

        constructor(the_object, *parameters)
        return the_object

    def __reflexively_inject_attributes(self, clazz, body):
        attributes = clazz.__annotations__
        the_object = object.__new__(clazz)
        for attribute, declared_type in attributes.items():
            allowed_attribute_types = [declared_type] if not isinstance(declared_type,
                                                                        _Union) else declared_type.__args__
            if isinstance(declared_type, TypingMeta) and declared_type.__name__ == 'List':
                the_object.__setattr__(attribute,
                                       [self.deserialize(body[attribute][x], declared_type.__args__[0]) for x in
                                        range(0, len(body[attribute]))])
            elif Stream(allowed_attribute_types).anyMatch(
                    lambda attribute_type: isinstance(body.get(attribute), attribute_type)):
                the_object.__setattr__(attribute, body.get(attribute))
            else:
                raise IncorrectAttributeTypeException(attribute, declared_type)
        return the_object



import yaml


class ApplicationProperties(object):

    def __init__(self, content: dict):
        self.content = content

    def __getitem__(self, item):
        return self.content[item]


def load_yaml_config(filename: str) -> ApplicationProperties:
    with open(filename, 'r') as stream:
        return ApplicationProperties(yaml.load(stream))



from jivago.inject.registry import Component, Singleton


@Component
class SomeBean(object):
    def say_hello(self):
        return "hello"


@Component
@Singleton
class SomeOtherBean(object):
    pass



from typing import Callable

from jivago.inject.registry import Annotation


@Annotation
def GET(wrapped: Callable) -> Callable:
    return wrapped


@Annotation
def POST(wrapped: Callable) -> Callable:
    return wrapped


@Annotation
def DELETE(wrapped: Callable) -> Callable:
    return wrapped


@Annotation
def PUT(wrapped: Callable) -> Callable:
    return wrapped


http_methods = [GET, POST, DELETE, PUT]
method_strings = {'GET': GET, 'POST': POST, 'DELETE': DELETE, 'PUT': PUT}


def to_method(method_name: str) -> Annotation:
    return method_strings[method_name]



from typing import Union, Callable


class Registration(object):
    def __init__(self, registered: Union[type, Callable], *, arguments=None):
        self.registered = registered
        self.arguments = arguments if arguments is not None else {}

    def is_in_package(self, package_name: str) -> bool:
        return self.registered.__module__.startswith(package_name)



from typing import List, Type

from jivago.config.production_jivago_context import ProductionJivagoContext
from jivago.inject.registry import Registry
from jivago.lang.annotations import Override
from jivago.wsgi.filters.debug_exception_filter import DebugExceptionFilter
from jivago.wsgi.filters.exception_filter import ExceptionFilter
from jivago.wsgi.filters.filter import Filter
from jivago.wsgi.filters.json_serialization_filter import JsonSerializationFilter


class DebugJivagoContext(ProductionJivagoContext):

    def __init__(self, root_package: str, registry: Registry):
        super().__init__(root_package, registry)

    @Override
    def get_filters(self, path: str) -> List[Type[Filter]]:
        return [ExceptionFilter, DebugExceptionFilter, JsonSerializationFilter]



from jivago.config.abstract_binder import AbstractBinder
from jivago.inject.registry import Registry, Provider
from jivago.inject.service_locator import ServiceLocator
from jivago.lang.annotations import Override
from jivago.lang.stream import Stream


class ProviderBinder(AbstractBinder):

    def __init__(self, root_package: str, registry: Registry):
        self.rootPackage = root_package
        self.registry = registry

    @Override
    def bind(self, service_locator: ServiceLocator):
        providers = self.registry.get_annotated_in_package(Provider, self.rootPackage)
        Stream(providers).map(lambda r: r.registered).forEach(lambda p: service_locator.bind(p.return_type(), p))



import traceback

from jivago.wsgi.filters.filter import Filter
from jivago.wsgi.request import Request
from jivago.wsgi.response import Response


class DebugExceptionFilter(Filter):

    def doFilter(self, request: Request, response: Response, chain: "FilterChain"):
        try:
            chain.doFilter(request, response)
        except Exception as e:
            response.status = 500
            response.body = "{}\n{}".format(e.__class__.__name__, traceback.format_exc())



from jivago.inject.registry import ParametrizedAnnotation


@ParametrizedAnnotation
def Resource(value: str):
    return lambda x: x


@ParametrizedAnnotation
def Path(value: str):
    return lambda x: x




import json

from jivago.wsgi.filters.filter import Filter
from jivago.wsgi.filters.filter_chain import FilterChain
from jivago.wsgi.request import Request
from jivago.wsgi.response import Response


class JsonSerializationFilter(Filter):

    def doFilter(self, request: Request, response: Response, chain: FilterChain):
        if request.headers['Content-Type'] == 'application/json':
            request.body = json.loads(request.body)

        chain.doFilter(request, response)

        if isinstance(response.body, dict):
            response.body = json.dumps(response.body)
            response.headers['Content-Type'] = 'application/json'



from typing import Callable

from jivago.inject.injectable import Injectable
from jivago.inject.registry import Annotation


def Override(fun: Callable) -> Callable:
    return fun


def Inject(fun: Callable) -> Callable:
    return Injectable(fun)


@Annotation
def Serializable(clazz: type) -> type:
    return clazz

@Annotation
def BackgroundWorker(clazz: type) -> type:
    return clazz



from typing import List

from jivago.wsgi.filters.filter import Filter
from jivago.wsgi.request import Request
from jivago.wsgi.resource_invocator import ResourceInvocator
from jivago.wsgi.response import Response


class FilterChain(object):

    def __init__(self, filters: List[Filter], resource_call_wrapper: ResourceInvocator):
        self.resource_call_wrapper = resource_call_wrapper
        self.filters = filters

    def doFilter(self, request: Request, response: Response):
        if len(self.filters) == 0:
            gotten_response = self.resource_call_wrapper.invoke(request)
            response.copy(gotten_response)
        else:
            self.filters[0].doFilter(request, response, self.getNextChain())

    def getNextChain(self):
        return FilterChain(self.filters[1::], self.resource_call_wrapper)



from jivago.config.abstract_context import AbstractContext
from jivago.inject.registry import Registry
from jivago.inject.service_locator import ServiceLocator
from jivago.lang.stream import Stream
from jivago.wsgi.annotations import Resource
from jivago.wsgi.dto_serialization_handler import DtoSerializationHandler
from jivago.wsgi.filters.filter_chain import FilterChain
from jivago.wsgi.request import Request
from jivago.wsgi.resource_invocator import ResourceInvocator
from jivago.wsgi.response import Response
from jivago.wsgi.routing_table import RoutingTable


class Router(object):
    def __init__(self, registry: Registry, rootPackage, service_locator: ServiceLocator, context: AbstractContext):
        self.context = context
        self.serviceLocator = service_locator
        self.registry = registry
        self.rootPackage = rootPackage
        self.routingTable = RoutingTable(registry,
                                         self.registry.get_annotated_in_package(Resource, self.rootPackage.__name__))
        self.resourceInvocator = ResourceInvocator(service_locator, self.routingTable,
                                                   DtoSerializationHandler(registry, self.rootPackage.__name__))

    def route(self, env, start_response):
        path = env['PATH_INFO']
        instantiated_filters = Stream(self.context.get_filters(path)).map(
            lambda clazz: self.serviceLocator.get(clazz)).toList()
        filter_chain = FilterChain(instantiated_filters, self.resourceInvocator)

        headers = Stream(env.items()).filter(lambda k, v: k.startswith("HTTP")).toDict()
        headers['CONTENT-TYPE'] = env.get('CONTENT_TYPE')
        request_size = int(env.get('CONTENT_LENGTH')) if 'CONTENT_LENGTH' in env else 0
        body = env['wsgi.input'].read(request_size)
        request = Request(env['REQUEST_METHOD'], path, headers, env['QUERY_STRING'], body)
        response = Response.empty()

        filter_chain.doFilter(request, response)

        start_response(self.__get_status_string(response.status), [x for x in response.headers.items()])
        return [response.body.encode('utf-8')]

    def __get_status_string(self, status: int) -> str:
        # TODO actual status formatting
        if status == 200:
            return "200 OK"
        else:
            return "{} status".format(status)



from typing import List

from jivago.inject.registry import Annotation
from jivago.wsgi.ambiguous_routing_exception import AmbiguousRoutingException
from jivago.wsgi.route_registration import RouteRegistration
from jivago.wsgi.unknown_path_exception import UnknownPathException

PATH_PARAMETER = '{param}'


class RouteNode(object):
    def __init__(self):
        self.children = {}
        self.invocators = {}

    def register_child(self, path: List[str], http_primitive: Annotation, route_registration: RouteRegistration):
        if len(path) == 0:
            if http_primitive in self.invocators:
                raise AmbiguousRoutingException(http_primitive, route_registration)
            self.invocators[http_primitive] = route_registration
        else:
            next_path_element = PATH_PARAMETER if path[0].startswith("{") and path[0].endswith('}') else path[0]
            if next_path_element not in self.children:
                self.children[next_path_element] = RouteNode()
            self.children[next_path_element].register_child(path[1::], http_primitive, route_registration)

    def explore(self, path: List[str]) -> "RouteNode":
        if len(path) == 0:
            return self

        next_path_element = path[0]
        if next_path_element in self.children:
            return self.children[next_path_element].explore(path[1::])
        elif PATH_PARAMETER in self.children:
            return self.children[PATH_PARAMETER].explore(path[1::])
        raise UnknownPathException()



from typing import Callable

from jivago.inject.exception.undefined_return_provider_function import UndefinedReturnProviderFunction


class ProviderFunction(object):

    def __init__(self, wrapped_function: Callable):
        self.wrappedFunction = wrapped_function

    def __call__(self, *args, **kwargs):
        return self.wrappedFunction(*args, **kwargs)

    def return_type(self) -> type:
        return_type = self.wrappedFunction.__annotations__.get('return')
        if return_type:
            return return_type
        else:
            raise UndefinedReturnProviderFunction(self.wrappedFunction)



from jivago.wsgi.filters.filter import Filter
from jivago.wsgi.filters.filter_chain import FilterChain
from jivago.wsgi.request import Request
from jivago.wsgi.response import Response


class ExceptionFilter(Filter):

    def doFilter(self, request: Request, response: Response, chain: FilterChain):
        try:
            chain.doFilter(request, response)
        except Exception:
            response.status = 500
            response.body = "An error occurred while handling this request."



from typing import Callable


class Injectable(object):

    def __init__(self, constructor_function: Callable):
        self.constructorFunction = constructor_function

    def __call__(self, *args):
        self.constructorFunction(*args)

    @property
    def __annotations__(self):
        return self.constructorFunction.__annotations__

    @property
    def __code__(self):
        return self.constructorFunction.__code__



from jivago.wsgi.headers import Headers


class Response(object):

    def __init__(self, status: int, headers: dict, body):
        self.status = status
        self.headers = Headers(headers)
        self.body = body

    def copy(self, response: "Response"):
        # For using the response as an output parameter
        self.status = response.status
        self.headers = response.headers
        self.body = response.body

    @staticmethod
    def empty():
        return Response(200, {}, "")



import urllib.parse

from jivago.inject.service_locator import ServiceLocator
from jivago.wsgi.dto_serialization_handler import DtoSerializationHandler
from jivago.wsgi.methods import to_method
from jivago.wsgi.request import Request
from jivago.wsgi.response import Response
from jivago.wsgi.routing_table import RoutingTable

ALLOWED_URL_PARAMETER_TYPES = [str, int, float]


class ResourceInvocator(object):
    def __init__(self, service_locator: ServiceLocator, routing_table: RoutingTable,
                 dto_serialization_handler: DtoSerializationHandler):
        self.dto_serialization_handler = dto_serialization_handler
        self.routing_table = routing_table
        self.service_locator = service_locator

    def invoke(self, request: Request) -> Response:
        method = to_method(request.method)
        route_registration = self.routing_table.get_route_registration(method, request.path)
        resource = self.service_locator.get(route_registration.resourceClass)
        path_parameters = route_registration.parse_path_parameters(request.path)
        query_parameters = request.parse_query_parameters()

        parameter_declaration = route_registration.routeFunction.__annotations__.items()
        parameters = []
        for name, clazz in parameter_declaration:
            if name == 'return':  # This is the output type annotation
                break
            if clazz == Request:
                parameters.append(request)
            elif clazz == dict:
                parameters.append(request.body)
            elif self.dto_serialization_handler.is_serializable(clazz):
                parameters.append(self.dto_serialization_handler.deserialize(request.body, clazz))
            elif clazz in ALLOWED_URL_PARAMETER_TYPES:
                if name in path_parameters:
                    parameters.append(clazz(self._url_parameter_unescape(path_parameters[name])))
                elif name in query_parameters:
                    parameters.append(clazz(self._url_parameter_unescape(query_parameters[name])))

        function_return = route_registration.routeFunction(resource, *parameters)

        if isinstance(function_return, Response):
            return function_return
        elif self.dto_serialization_handler.is_serializable(function_return.__class__):
            function_return = self.dto_serialization_handler.serialize(function_return)
        return Response(200, {}, function_return)

    def _url_parameter_unescape(self, escaped):
        if isinstance(escaped, str):
            return urllib.parse.unquote(escaped)
        else:
            return escaped



from typing import List, Type

from jivago.inject.service_locator import ServiceLocator
from jivago.wsgi.filters.filter import Filter


class AbstractContext(object):
    INSTANCE = None

    def __init__(self):
        self.serviceLocator = ServiceLocator()
        self.configure_service_locator()
        self.INSTANCE = self
        self.serviceLocator.bind(ServiceLocator, self.serviceLocator)

    def configure_service_locator(self):
        raise NotImplementedError

    def service_locator(self) -> ServiceLocator:
        return self.serviceLocator

    def get_filters(self, path: str) -> List[Type[Filter]]:
        raise NotImplementedError



from typing import List, Type

from jivago.config.abstract_context import AbstractContext
from jivago.inject.annoted_class_binder import AnnotatedClassBinder
from jivago.inject.provider_binder import ProviderBinder
from jivago.inject.registry import Singleton, Component, Registry
from jivago.inject.scope_cache import ScopeCache
from jivago.lang.annotations import Override, BackgroundWorker
from jivago.lang.stream import Stream
from jivago.wsgi.annotations import Resource
from jivago.wsgi.filters.exception_filter import ExceptionFilter
from jivago.wsgi.filters.filter import Filter
from jivago.wsgi.filters.json_serialization_filter import JsonSerializationFilter


class ProductionJivagoContext(AbstractContext):

    def __init__(self, root_package: str, registry: Registry):
        self.rootPackage = root_package
        self.registry = registry
        super().__init__()

    @Override
    def configure_service_locator(self):
        AnnotatedClassBinder(self.rootPackage, self.registry, Component).bind(self.serviceLocator)
        AnnotatedClassBinder(self.rootPackage, self.registry, Resource).bind(self.serviceLocator)
        AnnotatedClassBinder(self.rootPackage, self.registry, BackgroundWorker).bind(self.serviceLocator)
        ProviderBinder(self.rootPackage, self.registry).bind(self.serviceLocator)
        for scope in self.scopes():
            scoped_classes = Stream(self.registry.get_annotated_in_package(scope, self.rootPackage)).map(
                lambda registration: registration.registered).toList()
            cache = ScopeCache(scope, scoped_classes)
            self.serviceLocator.register_scope(cache)

        Stream(self.get_filters("")).forEach(lambda f: self.serviceLocator.bind(f, f))

    def scopes(self) -> List[type]:
        return [Singleton, BackgroundWorker]

    @Override
    def get_filters(self, path: str) -> List[Type[Filter]]:
        return [ExceptionFilter, JsonSerializationFilter]



from jivago.lang.stream import Stream
from jivago.wsgi.headers import Headers


class Request(object):
    def __init__(self, method: str, path: str, headers: dict, query_string: str, body):
        self.method = method
        self.path = path
        self.headers = Headers(headers)
        self.body = body
        self.queryString = query_string

    def parse_query_parameters(self) -> dict:
        pairs = self.queryString.split("&")
        return Stream(pairs).filter(lambda x: x != "").map(lambda pair: pair.split("=")).toDict()



from typing import Callable, List

from jivago.lang.stream import Stream


class RouteRegistration(object):
    def __init__(self, resource_class: type, route_function: Callable, registered_path: List[str]):
        self.resourceClass = resource_class
        self.routeFunction = route_function
        self.registeredPath = registered_path

    def parse_path_parameters(self, gotten_path: str) -> dict:
        path = Stream(gotten_path.split('/')).filter(lambda x: x != "").toList()

        parameters = {}
        for gotten, registered in zip(path, self.registeredPath):
            if self.__is_parameter(registered):
                parameters[self.__format_parameter_name(registered)] = gotten

        return parameters

    def __format_parameter_name(self, parameter_declaration: str) -> str:
        return parameter_declaration.replace('{', '').replace('}', '')

    def __is_parameter(self, registered: str) -> bool:
        return registered.startswith('{') and registered.endswith('}')



from typing import Callable, List

from jivago.inject.provider_function import ProviderFunction
from jivago.inject.registration import Registration
from jivago.lang.stream import Stream


class Registry(object):
    content = {}

    def get_annotated_in_package(self, annotation: "Annotation", package: str) -> List[Registration]:
        if annotation not in self.content:
            return []
        annotated = self.content[annotation]
        return Stream(annotated).filter(lambda r: r.is_in_package(package)).toList()

    def register(self, label: "Annotation", clazz: type, *, arguments=None):
        if label not in self.content:
            self.content[label] = []
        self.content[label].append(Registration(clazz, arguments=arguments))

    def is_annotated(self, object: object, annotation: "Annotation"):
        if annotation not in self.content:
            return False
        return object in map(lambda registration: registration.registered, self.content[annotation])


class Annotation(object):

    def __init__(self, decorator: Callable):
        self.decorator = decorator
        self.registry = Registry()

    def __call__(self, target):
        decorator_call = self.decorator(target)
        self.registry.register(self, decorator_call)
        return decorator_call

    def __repr__(self):
        return self.decorator.__name__


class ParametrizedAnnotation(Annotation):

    def __call__(self, *args, **kwargs):
        type_hints = self.decorator.__annotations__
        arguments = []
        for key in type_hints.keys():
            if key == 'return':
                continue
            if key == 'value' and len(args) == 1:
                arguments.append((key, args[0]))
            elif key in kwargs:
                arguments.append((key, kwargs[key]))
            else:
                raise MissingAnnotationParameterException(key)

        return SimpleSaveDecorator(self.registry, self, arguments=dict(arguments))


class SimpleSaveDecorator(object):

    def __init__(self, registry: Registry, saveTarget, arguments: dict):
        self.saveTarget = saveTarget
        self.registry = registry
        self.arguments = arguments

    def __call__(self, target):
        self.registry.register(self.saveTarget, target, arguments=self.arguments)
        return target


@Annotation
def Singleton(wrapped_class: type) -> type:
    return wrapped_class


@Annotation
def Component(wrapped_class: type) -> type:
    return wrapped_class


@Annotation
def Provider(wrapped_function: Callable) -> Callable:
    return ProviderFunction(wrapped_function)


class MissingAnnotationParameterException(Exception):
    pass



from itertools import chain
from typing import Iterable, Callable, Iterator, Optional


class Stream(object):

    def __init__(self, *iterables: Iterable):
        self.iterable = chain(*iterables)

    def map(self, fun: Callable) -> "Stream":
        if self.__should_expand(fun):
            return Stream(map(lambda tup: fun(*tup), self.iterable))
        else:
            return Stream(map(fun, self.iterable))

    def filter(self, fun: Callable) -> "Stream":
        if self.__should_expand(fun):
            return Stream(filter(lambda tup: fun(*tup), self.iterable))
        else:
            return Stream(filter(fun, self.iterable))

    def forEach(self, fun: Callable) -> None:
        if self.__should_expand(fun):
            for i in self:
                fun(*i)
        else:
            for i in self:
                fun(i)

    def anyMatch(self, fun: Callable) -> bool:
        return any(self.map(fun))

    def allMatch(self, fun: Callable) -> bool:
        return all(self.map(fun))

    def firstMatch(self, fun: Callable) -> Optional:
        if self.__should_expand(fun):
            for i in self:
                if fun(*i):
                    return i
        else:
            for i in self:
                if fun(i):
                    return i
        return None

    def noneMatch(self, fun: Callable) -> bool:
        return not self.allMatch(fun)

    def flat(self) -> "Stream":
        return Stream(chain(*self))

    def toList(self) -> list:
        return [x for x in self.iterable]

    def toSet(self) -> set:
        return set(self.toList())

    def toDict(self) -> dict:
        return dict(self.toList())

    def __should_expand(self, fun: Callable) -> bool:
        return fun.__code__.co_argcount > 1

    def __iter__(self) -> Iterator:
        return iter(self.iterable)



from os import path
from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()
setup(
    name='jivago',
    version_format='{tag}',
    setup_requires=['setuptools-git-version'],
    description='A Python framework for writing Java-esque REST applications.',
    long_description=long_description,
    long_description_content_type='text/markdown',

    url='https://github.com/keotl/jivago',

    author='Kento A. Lauzon',
    author_email='kento.lauzon@ligature.ca',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
    ],

    packages=find_packages(exclude=['contrib', 'docs', 'tests', 'example_app']),

    # install_requires=['peppercorn'],  # Optional
    # extras_require={  # Optional
    #     'dev': ['check-manifest'],
    #     'test': ['coverage'],
    # },
)



from jivago.lang.annotations import Serializable


@Serializable
class MyDto(object):
    name: str
    age: int

    def __init__(self, name: str, age: int):
        self.name = name
        self.age = age



from typing import Callable, List

from jivago.inject.registration import Registration
from jivago.inject.registry import Annotation, Registry
from jivago.lang.stream import Stream
from jivago.wsgi.annotations import Path
from jivago.wsgi.methods import http_methods
from jivago.wsgi.route_registration import RouteRegistration
from jivago.wsgi.route_node import RouteNode


class RoutingTable(object):

    def __init__(self, registry: Registry, resources: List[Registration]):
        self.routeRootNode = RouteNode()
        for resource in resources:
            for primitive in http_methods:
                routable_functions = registry.get_annotated_in_package(primitive, resource.registered.__module__)
                sub_paths = registry.get_annotated_in_package(Path, resource.registered.__module__)

                for route_function in routable_functions:
                    route_sub_path = Stream(sub_paths).firstMatch(lambda r: r.registered == route_function.registered)
                    if route_sub_path is None:
                        self.__register_route(resource, route_function.registered, primitive)
                    else:
                        self.__register_route(resource, route_function.registered, primitive,
                                              route_sub_path.arguments['value'])

    def __register_route(self, resource: Registration, function: Callable, primitive: Annotation, subpath: str = ""):
        path = resource.arguments['value'].split('/')
        if subpath is not None:
            path.extend(subpath.split('/'))
        path = Stream(path).filter(lambda s: s != "").toList()
        self.routeRootNode.register_child(path, primitive, RouteRegistration(resource.registered, function, path))

    def get_route_registration(self, http_primitive: Annotation, path: str) -> RouteRegistration:
        path_elements = Stream(path.split('/')).filter(lambda x: x != "").toList()
        route_node = self.routeRootNode.explore(path_elements)
        return route_node.invocators[http_primitive]



from jivago.config.abstract_context import AbstractContext

contexts = {}


def ApplicationContext(wrapped_class: type) -> type:
    global contexts
    contexts[wrapped_class.__name__.lower()] = wrapped_class
    return wrapped_class


class UnknownApplicationContextException(Exception):
    pass


def create_context(context_name: str) -> AbstractContext:
    context = contexts.get(context_name.lower())()
    if context is None:
        raise UnknownApplicationContextException(context_name)
    return context



from jivago.wsgi.request import Request
from jivago.wsgi.response import Response


class Filter(object):

    def doFilter(self, request: Request, response: Response, chain: "FilterChain"):
        raise NotImplementedError



import pkgutil
from threading import Thread
from typing import List, Type

from jivago.config.abstract_context import AbstractContext
from jivago.config.debug_jivago_context import DebugJivagoContext
from jivago.config.production_jivago_context import ProductionJivagoContext
from jivago.inject.registry import Registry, Annotation
from jivago.lang.annotations import BackgroundWorker
from jivago.lang.stream import Stream
from jivago.wsgi.router import Router


class JivagoApplication(object):

    def __init__(self, root_module, *, debug: bool = False, context: AbstractContext = None):
        self.registry = Registry()
        if context is None:
            self.context = DebugJivagoContext(root_module.__name__, self.registry) \
                if debug else ProductionJivagoContext(root_module.__name__, self.registry)
        else:
            self.context = context
        self.rootModule = root_module
        self.__import_package_recursive(root_module)
        self.context.configure_service_locator()
        self.serviceLocator = self.context.service_locator()
        self.router = Router(Registry(), self.rootModule, self.serviceLocator, self.context)

        self.backgroundWorkers = Stream(self.get_annotated(BackgroundWorker)).map(
            lambda clazz: self.serviceLocator.get(clazz)).map(lambda worker: Thread(target=worker))
        Stream(self.backgroundWorkers).forEach(lambda thread: thread.start())

    def __import_package_recursive(self, package):
        prefix = package.__name__ + "."
        for importer, modname, ispkg in pkgutil.iter_modules(package.__path__, prefix):
            module = __import__(modname, fromlist="dummy")
            if ispkg:
                self.__import_package_recursive(module)

    def get_annotated(self, annotation: Annotation) -> List[Type]:
        return Stream(Registry().get_annotated_in_package(annotation, self.rootModule.__name__)).map(
            lambda registration: registration.registered).toList()

    def __call__(self, env, start_response):
        """wsgi entry point."""
        return self.router.route(env, start_response)
